##############################################################################
# cgiforms.py Version 0.9.15
# (c) by Michael Stroeder <michael@stroeder.com>
##############################################################################
# This module is distributed under the terms of the
# GPL (GNU GENERAL PUBLIC LICENSE) Version 2
# (see http://www.gnu.org/copyleft/gpl.html)
##############################################################################
# Klassen zum Behandeln <FORM>s und zu CGI-BINs bermittelten Forminhalten
#
# Anmerkungen:
# - Eingabefelder mit gleichen name-Attributen werden logisch zusammengefasst
#   und muessen in der selben Reihenfolge registriert und angegeben werden.
# - 
# - Es werden nur Parameter aus dem QUERY_STRING (GET) oder einzelne
#   Datensaetze vom Typ application/x-www-form-urlencoded (POST) als
#   Parameter angenommen.
###############################################################################

import sys, os, types, string, re, urllib, urlparse, charset
from types import StringType

###############################################################################
# Klassen zur Verwaltung von einzelnen Form-Feldern
###############################################################################

###############################################################################
# Basisklasse fuer beliebige Eingabe-Felder.
# Sollte nicht direkt benutzt werden,
# spezielle Felder sollten von dieser Klasse abgeleitet werden 
###############################################################################

class formFieldClass:

  def __init__(self,name,text,maxlength=0,pattern='',default='',required=0,multiple=0):
    self.name      = name
    self.text      = text
    self.maxlength = maxlength
    if type(pattern) is types.TupleType:
      # pattern is tuple (regex,re options)
#      self.regex = re.compile('(%s)' % pattern[0],pattern[1])
      self.regex = re.compile('^%s$' % pattern[0],pattern[1])
    else:
      # pattern is string
#      self.regex = re.compile('(%s)' % pattern)
      self.regex = re.compile('^%s$' % pattern)
    self.default   = default
    self.required  = required
    self.multiple  = multiple
    self.charset   = 'iso-8859-1'
    self.counter   = 0
    self.inputfield_template = r'%s'
    self.contentprint_template = r'%s'
    if multiple:
      self.content   = []
    else:
      self.content   = ''

  # Feldinhalt setzen
  def setdefault(self,input):
    self.default = input

  # Feldinhalt setzen
  def put(self,input):
    if self.multiple and len(self.content)<self.multiple:
      if self.content:
        self.content.append(input)
      else:
        self.content = [input]
    else:
      self.content = input

  # Feldinhalt setzen
  def setcharset(self,charset):
    self.charset = charset

  # Simple Text-Ausgabe des Feldinhalts
  def textprint(self):
    return self.content
    
  # HTML-Ausgabe des Feldinhalts
  def defaultprint(self,escape_html_chars='<>":={}()'):
    return charset.recode(charset.escapeHTML(self.default,escape_html_chars),self.charset,'html4')

  # HTML-Ausgabe des Feldinhalts
  def contentprint(self,escape_html_chars='<>":={}()'):
    return self.contentprint_template % charset.recode(charset.escapeHTML(self.content,escape_html_chars),self.charset,'html4')

###############################################################################
# Klasse fuer <TEXTAREA>
###############################################################################

class formTextareaClass(formFieldClass):

  def __init__(self,name,text,maxlength=0,pattern='',default='',required=0,multiple=0,rows=10,cols=60):
    formFieldClass.__init__(self,name,text,maxlength,(pattern,re.M+re.S),default,required,multiple)
    self.rows  = rows
    self.cols  = cols

  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self):
    return '<textarea name="%s" rows=%d cols=%d>%s</textarea>' % \
           (self.name,self.rows,self.cols,self.defaultprint())

###############################################################################
# Klasse fuer <INPUT>
###############################################################################

class formInputClass(formFieldClass):

  def __init__(self,name,text,maxlength=0,pattern='',default='',required=0,multiple=0,size=0):
    formFieldClass.__init__(self,name,text,maxlength,pattern,default,required,multiple)
    if size:
      self.size = size
    else:
      self.size = maxlength

  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self):
    return '<input name="%s" maxlength=%d size=%d value="%s">' % (
      self.name,self.maxlength,self.size,self.defaultprint()
    )


###############################################################################
# Klasse fuer <INPUT TYPE=HIDDEN>
###############################################################################

class formHiddenInputClass(formInputClass):

  def __init__(self,name,text,maxlength=0,pattern='',default='',required=0,multiple=0,show=0):
    formFieldClass.__init__(self,name,text,maxlength,pattern,default,required,multiple)
    self.show = show

  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self):
    if self.show:
      return self.inputfield_template % (
               '<input type=hidden name="%s" value="%s">%s' % (
	         self.name,self.defaultprint(),self.defaultprint()
	       )
	     )
    else:    
      return self.inputfield_template % (
               '<input type=hidden name="%s" value="%s">' % (
	         self.name,self.defaultprint()
	       )
	     )


###############################################################################
# Klasse fuer <INPUT TYPE=password>
# ist von <INPUT> getrennte Klasse, da eigene contentprint-Methode
###############################################################################

class formPasswordClass(formFieldClass):

  def __init__(self,name,text,maxlength=0,pattern='',required=0,multiple=0,size=0):
    formFieldClass.__init__(self,name,text,maxlength,pattern,'',required)
    if size:
      self.size = size
    else:
      self.size = maxlength
  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self):
    return self.inputfield_template % (
             '<input name="%s" maxlength=%d size=%d type=password value="">' % (
	       self.name,self.maxlength,self.size
	     )
	   )
  # Simple Text-Ausgabe des Feldinhalts
  def textprint(self):
    return len(self.content)*'*'
      # HTML-Ausgabe des Feldinhalts
  def contentprint(self):
    return self.contentprint_template % (len(self.content)*'*')


###############################################################################
# Klasse fuer <SELECT>
###############################################################################

class formSelectClass(formFieldClass):

  def __init__(self,name,text,options=[],default='',required=0,multiple=0,size=1):

    if not (type(default) is types.ListType):
      default = [default]
    # pattern und maxlength werden aus den Optionen zusammengebaut
    if options:
      maxlength = 0
      valuelist=[]
      for i in options:
        if type(i) is types.TupleType:
  	  valuelist.append(i[0])
          if len(i[0])>maxlength:
	    maxlength=len(i[0])
	else:
  	  valuelist.append(i)
          if len(i)>maxlength:
	    maxlength=len(i)
      pattern   = '^%s$' % string.join(map(re.escape,valuelist),'$|^')
    else:
      pattern   = ''
      maxlength = 0
    if multiple:
      multiple = len(options)
    formFieldClass.__init__(self,name,text,maxlength,pattern,default,required,multiple)
    self.options  = options
    self.size     = size

  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self,escape_html_chars='<>":={}()'):
    s = ['<select name="%s" size=%d %s>' % (self.name,self.size," multiple"*(self.multiple>0))]
    for i in self.options:
      if type(i) is types.TupleType:
        optionvalue = i[0]
	optiontext = i[1]
      else:
        optionvalue = optiontext = i
      if type(self.default) is types.ListType:
	optionselected = optionvalue in self.default
      else:
	optionselected = optionvalue == self.default
      s.append(
	'<option value="%s"%s>%s</option>' % (
	  charset.recode(charset.escapeHTML(optionvalue,escape_html_chars),self.charset,'html4'),
	  ' selected'*(optionselected),
	  charset.recode(charset.escapeHTML(optiontext,escape_html_chars),self.charset,'html4')
	)
      )
    s.append('</select>')
    return self.inputfield_template % string.join(s,'\n')

###############################################################################
# Klasse fuer <INPUT TYPE=RADIO>
###############################################################################

class formRadioClass(formFieldClass):

  def __init__(self,name,text,options='',default='',required=0,multiple=0):
    # pattern und maxlength werden aus den Optionen zusammengebaut
    if len(options):
      pattern   = '(%s)' % string.join(map(re.escape,options),'|')
      maxlength = len(options[0])
      for i in options[1:]:
        if len(i)>maxlength:
	  maxlength=len(i)
    else:
      pattern   = ''
      maxlength = 0
    formFieldClass.__init__(self,name,text,maxlength,pattern,default,required,multiple)
    self.options  = options

  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self):
    s = []
    for i in self.options:
      s.append('<input type=radio name="%s" value="%s"%s>%s<br>' % (
                  self.name,i,' checked'*(i==self.default),i
		)
	      )
    return self.inputfield_template % string.join(s,'\n')

###############################################################################
# Klasse fuer <INPUT TYPE=CHECKBOX>
###############################################################################

class formCheckboxClass(formFieldClass):

  def __init__(self,name,text,value='',checked=0,required=0,multiple=0):
    pattern      = value
    maxlength    = len(value)
    formFieldClass.__init__(self,name,text,maxlength,pattern,'',required)
    self.value   = value
    self.checked = checked

  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self):
    return self.inputfield_template % (
             '<input type=checkbox name="%s" value="%s"%s>' % (
               self.name,self.value,' checked'*self.checked
	     )
	   )

###############################################################################
# Klasse fuer <KEYGEN>
###############################################################################

class formKeygenClass(formFieldClass):

  def __init__(self,name,text,maxlength=0,required=0,multiple=0):
    pattern      = r'[ -z\n]*'
    formFieldClass.__init__(self,name,text,maxlength,pattern,required)

  # Feldinhalt setzen
  def put(self,input):
    formFieldClass.put(self,string.translate(input, string.maketrans("",""),"\n"))

  # HTML-Ausgabe des Eingabefeldes
  def inputfield(self,challenge):
    return self.inputfield_template % (
             '<keygen name="%s" challenge="%s">' % (
               self.name,challenge
	     )
	   )

  # HTML-Ausgabe des Feldinhalts
  def contentprint(self):
    return self.contentprint_template % ('%d Bytes' % (len(self.content)))

#############################################################
# Klassen zur Verwaltung einer Ansammlung von Eingabefeldern
#############################################################


# Klassen zur Fehlerbehandlung von formClass
class formException(Exception):
  def __init__(self, *args):
      self.args = args


# Gesamtlaenge der eingegebenen Daten zu lang
class formContentLengthException(formException):
  def __init__(self,contentlength,maxcontentlength):
    formException.__init__(self,contentlength,maxcontentlength)
    self.contentlength = contentlength
    self.maxcontentlength = maxcontentlength
  def __str__(self):
    return 'Content length invalid. Expected at most %d bytes but received %d.\n' % (
      self.maxcontentlength,self.contentlength
    )


# Parameter mit unbekanntem Namen      
class formParamNameException(formException):
  def __init__(self,name):
    formException.__init__(self,name)
    self.name = name
  def __str__(self):
    return 'Unknown parameter %s.\n' % (self.name)


# Parameter mit unbekanntem Namen      
class formParamsMissing(formException):
  def __init__(self,paramnamelist):
    formException.__init__(self,paramnamelist)
    self.missing = paramnamelist
  def __str__(self):
    return 'Required fields missing: %s\n' % (
      string.join(self.missing,', ')
    )


# Inhalt eines Parameters ist ungueltig      
class formParamContentException(formException):
  def __init__(self,name,text,content,reqregex):
    formException.__init__(self,name,text,content,reqregex)
    self.name     = name
    self.text     = text
    self.content  = content
    self.reqregex = reqregex
  def __str__(self):
    return 'Content of field %s does not match "%s". Input was: "%s"\n' % (
      self.text,self.reqregex,self.content
    )


# Zuviele Parameter mit selbem Namen      
class formParamStructException(formException):
  def __init__(self,name,count,maxfields):
    formException.__init__(self,name,count,maxfields)
    self.name      = name
    self.count     = count
    self.maxfields = maxfields
  def __str__(self):
    return 'Expected at most %d parameters for field %s, got %d.\n' % (
      self.maxfields,self.name,self.count
    )


# Angegebener Parameter ist zu lang
class formParamLengthException(formException):
  def __init__(self,name,text,length,maxlength):
    formException.__init__(self,name,text,length,maxlength)
    self.name      = name
    self.text      = text
    self.length    = length
    self.maxlength = maxlength
  def __str__(self):
    return 'Content too long. Field %s has %d characters but is limited to %d.\n' % (
      self.text,self.length,self.maxlength
    )


#############################################################
# Die eigentlichen Klasse fuer ein <FORM>
#############################################################
      
class formClass:

  def __init__(
    self,
    inf = None,
    env = os.environ,
    charset = ''
  ):
    # Dictionary der Eingabefelder-Objekte
    # {name:[FormFieldClass]}
    self.field            = {}
    # Dictionary mit Ordnungszaehler der Eingabefelder-Objekte
    # {name:len(field)}
    self.fieldcounter     = {}
    # Reihenfolgetreue Liste der vorgesehenen Eingabefelder-Namen
    self.keys             = []
    # Liste der tatsaechlich eingegebenen Eingabefelder-Namen
    self.inputkeys        = []
    # Skriptname

    self.env = env

    if inf:
      self.inf = inf
    else:
      self.inf = sys.stdin

    self.request_method = env['REQUEST_METHOD']
    self.server_name = env.get('SERVER_NAME',env.get('HTTP_HOST',''))
    self.server_port = env.get('SERVER_PORT','')
    self.script_name = env['SCRIPT_NAME']
    self.path_info = env.get('PATH_INFO','')
    self.query_string = env.get('QUERY_STRING','')
    self.http_user_agent = env.get('HTTP_USER_AGENT','')
    if charset:
      self.accept_charset = charset
    else:
      self.accept_charset = string.split(env.get('HTTP_ACCEPT_CHARSET','utf-8'),',')[0]

    if self.request_method=='POST':
      # Parameter von self.inf lesen
      self.contentlength = int(env['CONTENT_LENGTH'])
      self.query_string = self.inf.read(self.contentlength)
    elif self.request_method=='GET':
      self.query_string = env.get('QUERY_STRING','')
      self.contentlength = len(self.query_string)
    else:
      raise ValueError,"Invalid request method %s." % self.request_method

    self.maxcontentlength = 0


  # Feld einem FORM hinzufuegen
  def add(self,formfield):
    formfield.setcharset(self.accept_charset)
    if self.field.has_key(formfield.name):
      formfield.counter = self.fieldcounter[formfield.name]
      self.field[formfield.name].append(formfield)
      self.fieldcounter[formfield.name] = self.fieldcounter[formfield.name] + 1
    else:
      formfield.counter = 0
      self.field[formfield.name] = [formfield]
      self.keys.append(formfield.name)
      self.fieldcounter[formfield.name] = 1
    self.maxcontentlength = self.maxcontentlength + formfield.maxlength

  # Alle Eingaben in inputform auf Gueltigkeit ueberpruefen.
  # Im Fehlerfall wird eine Ausnahme der Klasse formException generiert
  # Der Parameter ignoreemptyparams veranlat das Ignorieren leerer Eingaben.
  def getparams(self,ignoreemptyparams=0):

    # Parameterliste auseinandernehmen
    inputlist = string.split(self.query_string,'&')

    # Ueberhaupt Parameter angegeben?
    if not len(inputlist):
      return

    datalength = 0
    # Zaehlerdict. mit Index paramname
    paramnumindex = {}

    # Parameterliste abarbeiten
    for param in inputlist:

      if param:

        # Einzelne Parametername/-daten-Paare auseinandernehmen
        paramname,paramdata = string.split(param,'=',1)
	paramname = string.strip(urllib.unquote_plus(paramname))
	paramdata = string.strip(urllib.unquote_plus(paramdata))

        datalength = datalength+len(paramdata)

        # Gesamtlaenge der Daten noch zulaessig?
        if datalength > self.maxcontentlength:
          formContentLengthException(datalength)

        # Unbekannter Parametername angegeben?
	if not (paramname in self.keys):
          raise formParamNameException(paramname)

        if paramnumindex.has_key(paramname):
	  paramnumindex[paramname] = paramnumindex[paramname] + 1
	else:
	  paramnumindex[paramname] = 0

        # Anzahl der Parameter gleichen Namens noch zulaessig?
	if paramnumindex[paramname] >= len(self.field[paramname]):
          if (self.field[paramname][0].multiple):
  	    paramnumindex[paramname] = 0
          else:
            raise formParamStructException(paramname,paramnumindex[paramname]+1,len(self.field[paramname]))

        field = self.field[paramname][paramnumindex[paramname]]

        # kein Leerstring als Parameter?
	if paramdata:
          # Zusaetzlich pruefen
          # Laenge gueltig?
          if len(paramdata) > field.maxlength:
            raise formParamLengthException(field.name,field.text,len(paramdata),field.maxlength)
          rm = field.regex.match(paramdata)
          if rm==None or rm.group(0)!=paramdata:
            raise formParamContentException(field.name,field.text,paramdata,field.regex.pattern)
	  # Eingabe ist gueltig und wird in content uebernommen
          if not paramname in self.inputkeys:
	    self.inputkeys.append(paramname)
        else:
          if (not ignoreemptyparams) and (not paramname in self.inputkeys):
	    self.inputkeys.append(paramname)

        # Feldinhalt setzen
        field.put(paramdata)

    # Alle Parameter mit required-Flag vorhanden?
    missing_params = []
    for param in self.keys:
      for i in self.field[param]:
	if i.required and not ((param in self.inputkeys) and i.content):
	  missing_params.append(i.name,i.text)
    if missing_params:
      raise formParamsMissing(missing_params)

    return
